/*      program compare
 *
 * This program tests the H0-hypothesis that two time series are realizations
 * of the same process. It builds up phase space and calculates Q^ and Var{Q^}.
 * For identical processes S = Q^/Sqrt(Var{Q^}) has mean zero and standard
 * deviation one. For different processes larger values of S are expected.
 *
 */

#include <stdio.h>
#include <math.h>

#define FLOAT float
#define INT int
#define MAXN 100000

int main(void)
{
  char	 file[128];
  double x[MAXN], y[MAXN], dis, sigma, d0, d, Q, Qxx, Qyy, Qxy, g[2*MAXN], hlp,
	 av, av_sq, tmp, E_g, E_h, E_hh, E_gg, h_ij, var_Q;
  int    i, j, m, ntau, ndim, N1, N2, L, I, J;
  FILE   *infile;

/* read the time series */

  printf("File 1: "); scanf("%s", file);

  if ( (infile = fopen(file,"r")) == NULL)
  {
    fprintf(stderr,"Unable to open file\n");
    exit(1);
  }

  i = 0;
  while (fscanf(infile, "%lf", &x[i]) != EOF) i++;
  N1 = i;
  fclose(infile);
  fprintf(stderr,"%d data read\n", N1);

  printf("File 2: "); scanf("%s", file);

  if ( (infile = fopen(file,"r")) == NULL)
  {
    fprintf(stderr,"Unable to open file\n");
    exit(1);
  }

  i = 0;
  while (fscanf(infile, "%lf", &y[i]) != EOF) i++;
  N2 = i;
  fclose(infile);
  fprintf(stderr,"%d data read\n", N2);
  
  printf("delay (# sample times) : "); scanf("%d", &ntau);
  printf("embedding dim : "); scanf("%d", &ndim);
  printf("block length : "); scanf("%d", &L);
  printf("bandwidth : "); scanf("%lf", &d);
 
/* Determine mean and variance */

  av = 0.0;
  av_sq = 0.0;

  for(i=0;i!=N1;i++) {
    av += x[i];
    av_sq += x[i]*x[i];
  }

  for(i=0;i!=N2;i++) {
    av += y[i];
    av_sq += y[i]*y[i];
  }

  av /=  (N1+N2);
  av_sq /=  (N1+N2);
  

/*
 * renormalize the time series to have a standard deviation of 1/(2*sqrt(3))
 * A uniformly distributed time series gets a range of 1 in this way.
 * Both time series are scaled similar.
 */

  for(i=0;i!=N1;i++){
    x[i] = sqrt((double)(N1+N2)/(double)(N1+N2-1)) * (x[i]-av) / (2.0*sqrt(3.0*(av_sq-av*av)));
  }
  for(i=0;i!=N2;i++){
    y[i] = sqrt((double)(N1+N2)/(double)(N1+N2-1)) * (y[i]-av) / (2.0*sqrt(3.0*(av_sq-av*av)));
  }

/* switch to vectors */

  N1 -= (ndim-1)*ntau;
  N2 -= (ndim-1)*ntau;
  
  N1 = (int)((double)(N1)/L);
  N2 = (int)((double)(N2)/L);

  d0 = 2*d;


/* Initialize cumulative variables */

  Q = Qxx = Qyy = Qxy = 0.0;
  E_hh = 0.0;
  E_h = E_g = E_gg = 0.0;
  tmp = 0.0; 
  
  for (i=0;i!=N1+N2;i++) g[i] = 0.0;

  for (i=1;i!=N1+N2;i++) {
    for (j=0;j!=i;j++) {
      tmp = 0.0;
      h_ij = 0.0;
      if (i<N1 && j<N1) {
        for (I=i*L;I!=(i+1)*L;I++) {
	  for (J=j*L;J!=(j+1)*L;J++) {
	    dis = 0.0;
	    for (m=0;m!=ndim;m++) {
	      tmp = x[I+m*ntau] - x[J+m*ntau];
	      dis += tmp*tmp;
            }
            tmp = exp(-dis/(d0*d0));
            h_ij += tmp/(double)(L*L);
          } 
        }
        Qxx += h_ij;
      }
      else if (i>=N1 && j<N1) {  
        for (I=i*L;I!=(i+1)*L;I++) {
	  for (J=j*L;J!=(j+1)*L;J++) {
	    dis = 0.0;
	    for (m=0;m!=ndim;m++) {
	      tmp = y[I-N1*L+m*ntau] - x[J+m*ntau];
	      dis += tmp*tmp;
	    }
            tmp = exp(-dis/(d0*d0));
            h_ij += tmp/(double)(L*L);
          }
        }
        Qxy += h_ij;
      }
      else if (i>=N1 && j>=N1) {
        for (I=i*L;I!=(i+1)*L;I++) {
          for (J=j*L;J!=(j+1)*L;J++) {
            dis = 0.0;
            for (m=0;m!=ndim;m++) {
	      tmp = y[I-N1*L+m*ntau] - y[J-N1*L+m*ntau];
	      dis += tmp*tmp;
            } 
            tmp = exp(-dis/(d0*d0));
            h_ij += tmp/(double)(L*L);
          } 
        }
        Qyy += h_ij;
      }

      g[i] += h_ij;
      g[j] += h_ij;

      E_hh += h_ij*h_ij;
      E_h += h_ij;

    }
  }
 
  E_h /= (double)((N1+N2)*(N1+N2-1)/2.0);

  for (i=0;i!=N1+N2;i++)
    g[i] /= (double)(N1+N2-2);
    
/* determine the average E_g of the g_[i] */

/* in fact, this average can alternatively be
   expressed as E_g = E_h*(N1+N2-1)/(double)(N1+N2-2) */

  for (i=0;i!=N1+N2;i++) {
    E_g += g[i];
  }
  E_g /= (N1+N2);


/* center the g[i]'s and determine their mean square E_gg */

  for (i=0;i!=N1+N2;i++) {
    g[i] -= E_g;
    E_gg += g[i]*g[i];
  }
  E_gg /= (N1+N2);

/* normalize the mean square of the h's and the Q's found */

  E_hh /= (double)((N1+N2)*(N1+N2-1))/2.0;

  Qxx *= 2.0/(double)(N1*(N1-1));
  Qyy *= 2.0/(double)(N2*(N2-1));
  Qxy /= (double)(N1*N2);

  Q = Qxx + Qyy - 2.0*Qxy;

/* determine the conditional variance */


  var_Q = E_hh - E_h*E_h - 2.0*(N1+N2-2)*E_gg/(double)(N1+N2-1);

/*
 *
 *  hlp =			2.0/(double)(N1*(N1-1))  +
 *			2.0/(double)(N2*(N2-1))  +
 *			4.0/(double)(N1*N2);
 *
 *  hlp +=	-4.0*(
 *			(double)(N1-2)/(double)(N1*(N1-1)) +
 *		 	(double)(N2-2)/(double)(N2*(N2-1)) +	
 *			(double)(N1+N2-2)/(double)(N1*N2) -
 *			2.0*(1.0/(double)(N1) + 1.0/(double)(N2))
 *		    )*1.0/(double)(N1+N2-2);
 *  hlp += 
 *		    (
 *			(double)((N1-2)*(N1-3))/(double)(N1*(N1-1)) +
 *		 	(double)((N2-2)*(N2-3))/(double)(N2*(N2-1)) +
 *		 	4.0*(double)((N1-1)*(N2-1))/(double)(N1*N2) +
 *		 	2.0 +
 *		 	2.0*(double)(-2*(N1-2))/(double)(N1) +
 *		 	2.0*(double)(-2*(N2-2))/(double)(N2)
 *		    )*2.0/(double)((N1+N2-2)*(N1+N2-3));
 */

/*
 * This was the old way to calculate the 'hlp' variable The new way uses a
 * smaller expression found by rewriting the old expression :
 */  

  hlp = 2*(N1+N2-1)*(N1+N2-1)*(N1+N2-2);

  hlp /= (double)(N1+N2-3);   /* do'nt do them all at once, their product */
  hlp /= (double)(N1);        /* can well exceed MAXINT */   
  hlp /= (double)(N2);
  hlp /= (double)(N1-1);
  hlp /= (double)(N2-1);

  var_Q *= hlp;

  sigma = sqrt(var_Q);

  printf("Q = %12.9f\n", Q);
  printf("Qxx = %12.9f\n", Qxx);
  printf("Qyy = %12.9f\n", Qyy);
  printf("Qxy = %12.9f\n", Qxy);
  printf("sigma = %12.9f\n", sigma);
  printf("Q/sigma = %12.9f\n", Q/sigma);

  exit(0);
}

